"""
    us_lodes(conn::MySQL.Connection, selection::Dict) -> DataFrame

Retrieve annual job data for US regions using the LODES dataset from a MySQL database.

This function queries a MySQL database to retrieve annual job data for US regions based on the LODES dataset. It allows for customization of the query based on the `selection` dictionary, which specifies the year and geographic IDs to be included in the query.

# Arguments
- `conn::MySQL.Connection`: An active connection to a MySQL database.
- `selection::Dict`: A dictionary specifying the selection criteria including the year and geo_ids.

# Returns
- `DataFrame`: A DataFrame containing columns for the region ('US'), geographic ID, aggregation level, industry identifier, and average number of annual jobs.

# Example
> conn = MySQL.connect("localhost", "username", "password", db = "dbname")
> selection = Dict("year" => 2023, "geo_id" => ["US001", "US002"])
> df = us_lodes(conn, selection)

This example queries annual job data for US regions using the LODES dataset for the year 2023 and specific geographic IDs.
"""
function us_lodes(conn::MySQL.Connection, selection::Dict)
    year = selection["year"]
    geo_id = selection["geo_id"]
    if !isempty(geo_id)
        first_region_id_length = length(geo_id[1])
    else
        println("geo_id is empty")
    end
    dftemp = DataFrame()
    for id in geo_id
        if first_region_id_length == 15
            query = "SELECT sum(CNS01) as jobs_11, sum(CNS02) as jobs_21, sum(CNS03) as jobs_22, sum(CNS04) as jobs_23, sum(CNS05) as jobs_31_33, sum(CNS06) as jobs_42, sum(CNS07) as jobs_44_45, sum(CNS08) as jobs_48_49, sum(CNS09) as jobs_51, sum(CNS10) as jobs_52, sum(CNS11) as jobs_53, sum(CNS12) as jobs_54, sum(CNS13) as jobs_55, sum(CNS14) as jobs_56, sum(CNS15) as jobs_61, sum(CNS16) as jobs_62, sum(CNS17) as jobs_71, sum(CNS18) as jobs_72, sum(CNS19) as jobs_81, sum(CNS20) as jobs_92 FROM LODES8 WHERE Year = "*string(year)*" AND GeoID = '"*id*"';"
        elseif first_region_id_length < 15 && first_region_id_length > 0
            query = "SELECT sum(CNS01) as jobs_11, sum(CNS02) as jobs_21, sum(CNS03) as jobs_22, sum(CNS04) as jobs_23, sum(CNS05) as jobs_31_33, sum(CNS06) as jobs_42, sum(CNS07) as jobs_44_45, sum(CNS08) as jobs_48_49, sum(CNS09) as jobs_51, sum(CNS10) as jobs_52, sum(CNS11) as jobs_53, sum(CNS12) as jobs_54, sum(CNS13) as jobs_55, sum(CNS14) as jobs_56, sum(CNS15) as jobs_61, sum(CNS16) as jobs_62, sum(CNS17) as jobs_71, sum(CNS18) as jobs_72, sum(CNS19) as jobs_81, sum(CNS20) as jobs_92 FROM LODES8 WHERE Year = "*string(year)*" AND GeoID LIKE '"*id*"%';"
        end
        result = query_connection(conn, query)
        long_df = stack(result, names(result), variable_name=:industry, value_name=:jobs)
        long_df.industry = replace.(string.(long_df.industry), "jobs_" => "")
        long_df[!, :geo_id] = fill(id, nrow(long_df))
        long_df[!, :agg_level] = fill(2, nrow(long_df)) 
        dftemp = vcat(dftemp, long_df)
    end 
    #dftemp[!, :region] = fill("US", nrow(dftemp))
    return dftemp
end

"""
    us_qcew(conn::MySQL.Connection, selection::Dict) -> DataFrame

Retrieve annual job data for US regions using the QCEW dataset from a MySQL database.

This function queries a MySQL database to retrieve annual job data for US regions based on the QCEW (Quarterly Census of Employment and Wages) dataset. It allows for customization of the query based on the `selection` dictionary, which specifies the year, geographic IDs, and aggregation level to be included in the query.

# Arguments
- `conn::MySQL.Connection`: An active connection to a MySQL database.
- `selection::Dict`: A dictionary specifying the selection criteria including the year, geo_ids, and aggregation level.

# Returns
- `DataFrame`: A DataFrame containing columns for the region ('US'), geographic ID, aggregation level, industry identifier, and average number of annual jobs.

# Example
> conn = MySQL.connect("localhost", "username", "password", db = "dbname")
> selection = Dict("year" => 2023, "geo_id" => ["US001", "US002"], "agg_level" => 2)
> df = us_qcew(conn, selection)

This example queries annual job data for US regions using the QCEW dataset for the year 2023, specific geographic IDs, and aggregation level.
"""
function us_qcew(conn::MySQL.Connection, selection::Dict)
    # getting valuable info from dictionary
    year = selection["year"]
    geo_id = selection["geo_id"]
    # the agg level refers to the number of NAICS digits and goes from 2 to 6
    agg_level = selection["agg_level"]
    # object to be returned
    dftemp = DataFrame()
    # iterate over geo objects
    for id in geo_id
        # check what spatial level
        if id[1:2] == "US"
            agglvl_code = 10
        elseif parse(Int, id) < 100 || parse(Int, id) % 1000 == 0
            agglvl_code = 50
        else
            agglvl_code = 70
        end
        # infer Agglvl_code -- because of Agglvl_code codification:
        agglvl_code += (agg_level+2)
        # fix id for states
        if length(id) == 2
            full_id = string(id, "000")
        else
            full_id = id
        end

        # prepare query
        query = "SELECT Naics as industry, (Jan_jobs+Feb_jobs+Mar_jobs+Apr_jobs+May_jobs+Jun_jobs+Jul_jobs+Aug_jobs+Sep_jobs+Oct_jobs+Nov_jobs+Dec_jobs)/12 as jobs FROM QCEW WHERE Year = $year AND GeoID = '$full_id' AND Agglvl_code = '$agglvl_code';"
        # execute
        result = query_connection(conn, query)
        # Check if the 'industry' column exists
        if !("industry" in names(result))
             println("No data returned for query with GeoID = $full_id and Year = $year and AND Agglvl_code = $agglvl_code.")
            # Initialize `result` with the expected columns but no rows
            result = DataFrame(industry = String[], jobs = Float64[], geo_id = String[], agg_level = Int[])
        end
        # add geo_id col
        result[!, :geo_id] = fill(id, nrow(result))
        # add agg level
        result[!, :agg_level] = fill(agg_level, nrow(result))
        # copy results to be returned
        dftemp = vcat(dftemp, result)
    end 
    #dftemp[!, :region] = fill("US", nrow(dftemp))
    return dftemp
end

"""
    eu_lfs(conn::MySQL.Connection, selection::Dict) -> DataFrame
    # Description
    Retrieve annual job data for EU regions using the LFS dataset from a MySQL database.

    # Arguments
    - `conn::MySQL.Connection`: An active connection to a MySQL database.
    - `selection::Dict`: A dictionary specifying the selection criteria including the year and geo_ids.

    # Returns
    - `DataFrame`: A DataFrame containing columns for the region ('EU'), geographic ID, aggregation level, industry identifier, and average number of annual jobs.
"""
function eu_lfs(conn::MySQL.Connection, selection::Dict)
    # getting valuable info from dictionary
    year = selection["year"]
    geo_id = selection["geo_id"]
    # object to be returned
    dftemp = DataFrame()
    # iterate over geo objects
    for id in geo_id
        # prepare query
        query = "SELECT Nace as industry, ((EmpTh_Q1+EmpTh_Q2+EmpTh_Q3+EmpTh_Q4)/4)*1000 as jobs FROM LFS WHERE Year = $year AND GeoID = '$id';"
        # execute
        result = query_connection(conn, query)
        # add geo_id col
        result[!, :geo_id] = fill(id, nrow(result))
        # add agg level
        result[!, :agg_level] = fill(2, nrow(result))
        # copy results to be returned
        dftemp = vcat(dftemp, result)
    end 
    #dftemp[!, :region] = fill("EU", nrow(dftemp))
    return dftemp
end

"""
    eu_sbs(conn::MySQL.Connection, selection::Dict) -> DataFrame
    # Description
    Retrieve annual job data for EU regions using the SBS dataset from a MySQL database.

    # Arguments
    - `conn::MySQL.Connection`: An active connection to a MySQL database.
    - `selection::Dict`: A dictionary specifying the selection criteria including the year and geo_ids.

    # Returns
    - `DataFrame`: A DataFrame containing columns for the region ('EU'), geographic ID, aggregation level, industry identifier, and average number of annual jobs.
"""
function eu_sbs(conn::MySQL.Connection, selection::Dict)
        # getting valuable info from dictionary
        year = selection["year"]
        geo_id = selection["geo_id"]
        # object to be returned
        dftemp = DataFrame()
        # iterate over geo objects
        for id in geo_id
            # prepare query
            query = "SELECT Nace as industry, Employment as jobs FROM SBS WHERE Year = "*string(year)*" AND GeoID = '"*id*"';"
            # execute
            result = query_connection(conn, query)
            # add geo_id col
            result[!, :geo_id] = fill(id, nrow(result))
            # add agg level
            result[!, :agg_level] = fill(4, nrow(result))
            # copy results to be returned
            dftemp = vcat(dftemp, result)
        end 
        #dftemp[!, :region] = fill("EU", nrow(dftemp))
        return dftemp
end

"""
    eu_rea(conn::MySQL.Connection, selection::Dict) -> DataFrame
    # Description
    Retrieve annual job data for EU regions using the REA dataset from a MySQL database.

    # Arguments
    - `conn::MySQL.Connection`: An active connection to a MySQL database.
    - `selection::Dict`: A dictionary specifying the selection criteria including the year and geo_ids.

    # Returns
    - `DataFrame`: A DataFrame containing columns for the region ('EU'), geographic ID, aggregation level, industry identifier, and average number of annual jobs.
"""
function eu_rea(conn::MySQL.Connection, selection::Dict)
    # getting valuable info from dictionary
    year = selection["year"]
    geo_id = selection["geo_id"]
    # object to be returned
    dftemp = DataFrame()
    # iterate over geo objects
    for id in geo_id
        # prepare query
        query = "SELECT Nace as industry, EmpTh*1000 as jobs FROM REA WHERE Year = "*string(year)*" AND GeoID = '"*id*"';"
        # execute
        result = query_connection(conn, query)
        # add geo_id col
        result[!, :geo_id] = fill(id, nrow(result))
        # add agg level
        result[!, :agg_level] = fill(1, nrow(result))
        # copy results to be returned
        dftemp = vcat(dftemp, result)
    end 
    #dftemp[!, :region] = fill("EU", nrow(dftemp))
    return dftemp
end

"""
    get_annual_jobs(table::String, selection::Dict; host::String="127.0.0.1") -> DataFrame

Retrieve annual job data from a specific dataset and table in a MySQL database.

This function allows for the retrieval of annual job data based on the specified dataset and table name from a MySQL database. It provides flexibility by accepting a selection dictionary to customize the query criteria.

# Arguments
- `table::String`: The name of the dataset and table to query.
- `selection::Dict`: A dictionary specifying the selection criteria for the query, including the year, geo_ids, and aggregation level.
- `host::String` (optional): The host address for the MySQL database. Default is "127.0.0.1".

# Returns
- `DataFrame`: A DataFrame containing columns for the region ('US' or 'EU'), geographic ID, aggregation level, industry identifier, and average number of annual jobs.

# Example
> df = get_annual_jobs("us_lodes", Dict("year" => 2023, "geo_id" => ["US001", "US002"]), host="localhost")

This example queries annual job data for US regions using the LODES dataset for the year 2023 and specific geographic IDs.
"""
function get_annual_jobs(table::String, selection::Dict; host::String="127.0.0.1")
    table_dict = Dict(
        "us_lodes" => us_lodes,
        "us_qcew" => us_qcew,
        "eu_lfs" => eu_lfs,
        "eu_sbs" => eu_sbs,
        "eu_rea" => eu_rea
    )

    df_template = DataFrame(
        #region = String[], # column to store if EU or US
        geo_id = String[],  # Column for Geo IDs
        agg_level = Int64[], # Level of industrial classification
        industry = String[], # industry identifier
        jobs = Float64[] # column for average number of annual jobs
    )

    # Check if the table is supported
    if !haskey(table_dict, table)
        error("Unsupported table: $table. Please check the available tables: $(keys(table_dict)).")
    end
        
    # data retrieval
    df = get_selected_objects_from_table(table, df_template, table_dict, "FACT_jobs", selection; host=host, port=3308) 

    return df
end

